#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>

void print(double* M, int n0, int n1);

int main(int argc, char* argv[])
{
  MPI_Comm comm = MPI_COMM_WORLD;
  MPI_Init(&argc, &argv);

  int size, rank;
  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &rank);

  // initialize dimension n on root and broadcast to all processors
  int n;
  if (rank == 0)
  {
    if (argc != 2) MPI_Abort(comm, -1);
    n = atoi(argv[1]);
    if (n == 0) MPI_Abort(comm, -1);
  }
  MPI_Bcast(&n, 1, MPI_INT, 0, comm);

  // compute n0 such that size*n0 is the smallest multiple of size greater equal n
  int n0 = n%size == 0 ? n/size : 1+n/size;

  // initialize matrix A on root and scatter it into matrix A0 of every processor
  // row number n of A is extended to size*n0 such that A0 has n0 rows
  double *A;
  if (rank == 0)
  {
    A = malloc(size*n0*n*sizeof(double));
    for (int i=0; i<n; i++)
      for (int j=0; j<n; j++)
        A[i*n+j] = rand()%10;
    print(A, n, n);
  }
  double* A0 = malloc(n0*n*sizeof(double));
  MPI_Scatter(A, n0*n, MPI_DOUBLE, A0, n0*n, MPI_DOUBLE, 0, comm);
  if (rank == 0) free(A);

  // initialize B with n rows on root and broadcast it to all processors
  double* B = malloc(n*n*sizeof(double));
  if (rank == 0)
  {
    for (int i=0; i<n; i++)
      for (int j=0; j<n; j++)
        B[i*n+j] = rand()%10;
    print(B, n, n);
  }
  MPI_Bcast(B, n*n, MPI_DOUBLE, 0, comm);

  // compute C0 = A0*B with n0 rows on all processors
  double* C0 = malloc(n0*n*sizeof(double));
  for (int i=0; i<n0; i++)
  {
    for (int j=0; j<n; j++)
    {
      C0[i*n+j] = 0;
      for (int k=0; k<n; k++)
        C0[i*n+j] += A0[i*n+k]*B[k*n+j];
    }
  }
  free(A0); free(B);

  // gather on root C0 with size*n0 rows from all processors
  double* C;
  if (rank == 0) C = malloc(size*n0*n*sizeof(double));
  MPI_Gather(C0, n0*n, MPI_DOUBLE, C, n0*n, MPI_DOUBLE, 0, comm);
  free(C0);
  if (rank == 0) { print(C, n, n); free(C); }

  MPI_Finalize();
  return 0;
}

void print(double* M, int n0, int n1)
{
  printf("\n");
  for (int i=0; i<n0; i++)
  {
    for (int j=0; j<n1; j++)
      printf("%.0f ", M[i*n0+j]);
    printf("\n");
  }
}
